package io.gitlab.jfronny.commons.serialize;

import io.gitlab.jfronny.commons.SamWithReceiver;

import java.io.Flushable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.Pattern;

public abstract class SerializeWriter<TEx extends Exception, T extends SerializeWriter<TEx, T>> implements AutoCloseable, Flushable {
    private static final Pattern VALID_JSON_NUMBER_PATTERN =
            Pattern.compile("-?(?:0|[1-9][0-9]*)(?:\\.[0-9]+)?(?:[eE][-+]?[0-9]+)?");

    protected boolean lenient = false;
    protected boolean serializeNulls = true;
    protected boolean serializeSpecialFloatingPointValues = false;

    public boolean isLenient() {
        return lenient;
    }

    public T setLenient(boolean lenient) {
        this.lenient = lenient;
        if (lenient) return setSerializeSpecialFloatingPointValues(true).setSerializeNulls(true);
        return (T) this;
    }
    
    public boolean isSerializeNulls() {
        return serializeNulls;
    }
    
    public T setSerializeNulls(boolean serializeNulls) {
        this.serializeNulls = serializeNulls;
        return (T) this;
    }

    public boolean isSerializeSpecialFloatingPointValues() {
        return serializeSpecialFloatingPointValues;
    }
    
    public T setSerializeSpecialFloatingPointValues(boolean serializeSpecialFloatingPointValues) {
        this.serializeSpecialFloatingPointValues = serializeSpecialFloatingPointValues;
        return (T) this;
    }

    public abstract T beginArray() throws TEx;
    public abstract T endArray() throws TEx;
    public T array(SerializeWriterConsumer<TEx, T> consumer) throws TEx {
        return consumer.accept(this.beginArray()).endObject();
    }
    public abstract T beginObject() throws TEx;
    public abstract T endObject() throws TEx;
    public T object(SerializeWriterConsumer<TEx, T> consumer) throws TEx {
        return consumer.accept(this.beginObject()).endObject();
    }

    public abstract T comment(String comment) throws TEx;
    public abstract T name(String name) throws TEx;
    public T nullValue() throws TEx {
        if (serializeNulls) {
            return literalValue("null");
        } else {
            throw new IllegalArgumentException("Null values are not allowed");
        }
    }
    public abstract T value(String value) throws TEx;
    public T value(boolean value) throws TEx {
        return literalValue(value ? "true" : "false");
    }
    public T value(Boolean value) throws TEx {
        return value == null ? nullValue() : value(value.booleanValue());
    }
    public T value(float value) throws TEx {
        if (!serializeSpecialFloatingPointValues && (Float.isNaN(value) || Float.isInfinite(value)))
            throw new IllegalArgumentException("Numeric values must be finite, but was " + value);
        return literalValue(Float.toString(value));
    }
    public T value(Float value) throws TEx {
        return value == null ? nullValue() : value(value.floatValue());
    }
    public T value(double value) throws TEx {
        if (!serializeSpecialFloatingPointValues && (Double.isNaN(value) || Double.isInfinite(value)))
            throw new IllegalArgumentException("Numeric values must be finite, but was " + value);
        return literalValue(Double.toString(value));
    }
    public T value(Double value) throws TEx {
        return value == null ? nullValue() : value(value.doubleValue());
    }
    public T value(long value) throws TEx {
        return literalValue(Long.toString(value));
    }
    public T value(Long value) throws TEx {
        return value == null ? nullValue() : value(value.longValue());
    }
    public T value(Number value) throws TEx {
        if (value == null) return nullValue();
        String s = value.toString();
        if (s.equals("NaN") || s.equals("Infinity") || s.equals("-Infinity")) {
            if (!serializeSpecialFloatingPointValues) {
                throw new IllegalArgumentException("Numeric values must be finite, but was " + s);
            }
        } else {
            Class<? extends Number> numberClass = value.getClass();
            if (!isTrustedNumberClass(numberClass) && !VALID_JSON_NUMBER_PATTERN.matcher(s).matches()) {
                throw new IllegalArgumentException("String created by " + numberClass + " is not a valid number: " + s);
            }
        }
        return literalValue(s);
    }
    
    private static boolean isTrustedNumberClass(Class<? extends Number> c) {
        return c == Integer.class
                || c == Long.class
                || c == Double.class
                || c == Float.class
                || c == Byte.class
                || c == Short.class
                || c == BigDecimal.class
                || c == BigInteger.class
                || c == AtomicInteger.class
                || c == AtomicLong.class;
    }

    /**
     * Writes a literal value to the output without quoting or escaping.
     * This may not be supported by all implementations, if not supported an {@link UnsupportedOperationException} will be thrown.
     * @param value the literal value to write
     * @return this writer
     */
    public abstract T literalValue(String value) throws TEx;

    @Override
    public abstract void close() throws TEx;

    @SamWithReceiver
    public interface SerializeWriterConsumer<TEx extends Exception, T extends SerializeWriter<TEx, T>> {
        T accept(T writer) throws TEx;
    }
}
