package io.gitlab.jfronny.commons.throwable;

import java.util.function.Supplier;

/**
 * A utility class to hide checked exceptions from the java compiler.
 */
public class Assume {
    /**
     * Hides an exception from the java compiler, effectively making it unchecked.
     * Wraps the runnable in a runnable not marked to throw the checked exception.
     *
     * @param runnable the runnable to wrap
     * @param <TEx>    the exception the runnable might throw
     * @return         a runnable not marked to throw the checked exception
     */
    public static <TEx extends Throwable> Runnable isSafe(ThrowingRunnable<TEx> runnable) {
        var rn = (ThrowingRunnable<RuntimeException>) (ThrowingRunnable) runnable;
        return rn::run;
    }

    /**
     * Hides an exception from the java compiler, effectively making it unchecked.
     * Wraps the supplier in a supplier not marked to throw the checked exception.
     *
     * @param supplier the supplier to wrap
     * @param <T>      the return type of the supplier
     * @param <TEx>    the exception the supplier might throw
     * @return         a supplier not marked to throw the checked exception
     */
    public static <T, TEx extends Throwable> Supplier<T> isSafe(ThrowingSupplier<T, TEx> supplier) {
        var rn = (ThrowingSupplier<T, RuntimeException>) (ThrowingSupplier) supplier;
        return rn::get;
    }

    /**
     * Hides an exception from the java compiler, effectively making it unchecked.
     * Runs a runnable.
     *
     * @param runnable the runnable to run
     * @param <TEx>    the exception the runnable might throw
     */
    public static <TEx extends Throwable> void sneaky(ThrowingRunnable<TEx> runnable) {
        isSafe(runnable).run();
    }

    /**
     * Hides an exception from the java compiler, effectively making it unchecked.
     * Runs a supplier and returns the result.
     *
     * @param supplier the supplier to run
     * @param <T>      the return type of the supplier
     * @param <TEx>    the exception the supplier might throw
     * @return         the result of the supplier
     */
    public static <T, TEx extends Throwable> T sneaky(ThrowingSupplier<T, TEx> supplier) {
        return isSafe(supplier).get();
    }

    /**
     * Reintroduces an exception previously made sneaky, allowing you to catch it.
     * The call does nothing except tricking the compiler into thinking that the exception is thrown.
     *
     * @param <TEx> the exception to reintroduce into the set of possible exceptions
     * @throws TEx  the introduced exception
     */
    public static <TEx extends Throwable> void reintroduce() throws TEx {
    }

    /**
     * Reintroduces an exception previously made sneaky, allowing you to catch it.
     * Simply runs the runnable.
     *
     * @param runnable the runnable to run
     * @param <TEx>    the exception to reintroduce into the set of possible exceptions
     * @throws TEx     the exception that was thrown by the runnable
     */
    public static <TEx extends Throwable> void reintroduce(Runnable runnable) throws TEx {
        runnable.run();
    }

    /**
     * Reintroduces an exception previously made sneaky, allowing you to catch it.
     * Simply runs the supplier and returns the result.
     *
     * @param supplier the supplier to run
     * @param <T>      the return type of the supplier
     * @param <TEx>    the exception to reintroduce into the set of possible exceptions
     * @return         the result of the supplier
     * @throws TEx     the exception that was thrown by the supplier
     */
    public static <T, TEx extends Throwable> T reintroduce(Supplier<T> supplier) throws TEx {
        return supplier.get();
    }
}
